from sfaira.versions.metadata.base import Ontology, OntologyList, OntologyHierarchical, OntologyObo, \
    OntologyOboCustom, OntologyCl, OntologyHancestro, OntologyUberon, OntologyHsapdv, OntologyMondo, \
    OntologyMmusdv, OntologySex, OntologyEfo, OntologyCellosaurus, OntologyTaxon, OntologyUberonLifecyclestage
from sfaira.versions.metadata.universe import CelltypeUniverse
