from sfaira.versions.topologies.musmusculus.embedding.ae import AE_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.linear import LINEAR_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.nmf import NMF_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.vae import VAE_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.vaeiaf import VAEIAF_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.vaevamp import VAEVAMP_TOPOLOGIES
