"""
The classes contains constants related to sfaira streamlined meta data files.
"""

META_DATA_FIELDS = {
    "annotated": bool,
    "author": str,
    "assay_sc": str,
    "assay_differentiation": str,
    "assay_type_differentiation": str,
    "bio_sample": str,
    "cell_line": str,
    "cellontology_class": str,
    "cellontology_id": str,
    "development_stage": str,
    "disease": str,
    "doi": str,
    "download_url_data": str,
    "download_url_meta": str,
    "id": str,
    "ncells": int,
    "normalization": str,
    "organ": str,
    "organism": str,
    "sample_source": str,
    "state_exact": str,
    "tech_sample": str,
    "year": int,
}
