RELEASE_HUMAN = "104"
RELEASE_MOUSE = "104"
ASSEMBLY_HUMAN = f"Homo_sapiens.GRCh38.{RELEASE_HUMAN}"
ASSEMBLY_MOUSE = f"Mus_musculus.GRCm39.{RELEASE_MOUSE}"

CELLTYPES = ["adventitial cell", "endothelial cell", "acinar cell", "pancreatic PP cell", "type B pancreatic cell"]
CL_VERSION = "v2021-08-10"
