from sfaira.data.store.io.load_store import load_store
from sfaira.data.store.stores import StoreSingleFeatureSpace, \
    StoreAnndata, StoreDao, \
    StoreMultipleFeatureSpaceBase, \
    StoresAnndata, StoresDao, StoresH5ad
